//---------------------------------------------------------------------------
// JoySwitch v0.2 ALPHA, Joystick.cpp
// Copyright (C) 2003 by Donovan Hawkins <hawkins@cephira.com>
//
// This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//---------------------------------------------------------------------------

#include <vector>        // vector
#include <string>        // string
#include <objbase.h>     // CoInitializeEx, CoUninitialize, FAILED macro
#include "dinput.h"      // DirectInputCreate, LPDIRECTINPUT
#include "dinputd.h"     // DirectInputCreate, LPDIRECTINPUT (had to replace PHKEY with ::PHKEY for Builder)

#include "Joystick.h"


// Dummy GUID that should never match a joystick (generated just for this purpose)
// {7585CD8F-2398-452e-829E-073FE4E25047}

const GUID dummy = { 0x7585cd8f, 0x2398, 0x452e, { 0x82, 0x9e, 0x7, 0x3f, 0xe4, 0xe2, 0x50, 0x47 } };


// Call-back for DirectInput enumeration

BOOL CALLBACK  JoystickCallback( const DIDEVICEINSTANCE*  dinst, void*  joy )
{
  TJoySwitch*  j = static_cast< TJoySwitch* >( joy );

  for ( unsigned  i = 0; i < j->joystick.size(); ++i )  // have to look everywhere...DInput enum order is alphabetical
  {
    if ( j->joystick[ i ].ID == dinst->guidInstance )  // we'll try showing ALL matches, but should only be one
    {
      j->joystick[ i ].Name  = dinst->tszInstanceName;
      j->joystick[ i ].Found = true;
    }
  }

  return  true;
}


// Function that attempts to initialize everything (initialize di and dic to 0 before first call)

void  TJoySwitch::initialize()
{
  error = errNone;

 // Items that are initialized as part of DInput initialization
  if ( !di )
  {
   // initialize is_nt before proceeding
    OSVERSIONINFO  version;
    version.dwOSVersionInfoSize = sizeof( version );
    if ( !GetVersionEx( &version ) )
    {
      error = errWinVersion;
      return;
    }
    is_nt = ( version.dwPlatformId == VER_PLATFORM_WIN32_NT );
   // Arrange for di to be 0 if neither COM nor DInput are initialized, else both are initialized
    if ( FAILED( CoInitializeEx( NULL, COINIT_APARTMENTTHREADED ) ) )
    {
      error = errCOMInit;
      return;
    }
    if ( FAILED( DirectInputCreate( hinstance, DIRECTINPUT_VERSION, &di, NULL) ) )
    {
      error = errDInputInit;
      di = 0;
      CoUninitialize();
      return;
    }
  }

 // Items that are initialized as part of DInputJoyConfig initialization
  if ( !dic )
  {
   // Arrange for dic to be 0 if it isn't initialized or could not set coop or acquire, else all is ready
    if ( FAILED( di->QueryInterface( IID_IDirectInputJoyConfig, ( LPVOID* ) &dic ) ) )
    {
      error = errDInputJoyConfigInit;
      dic = 0;
      return;
    }
    if ( FAILED( dic->SetCooperativeLevel( hwnd, DISCL_EXCLUSIVE | DISCL_BACKGROUND ) ) )
    {
      error = errCoopLevel;
      dic->Release();
      dic = 0;
      return;
    }
//    HWND  oldhwnd = GetForegroundWindow();
    SetForegroundWindow( hwnd );
    HRESULT  result = dic->Acquire();
//    SetForegroundWindow( oldhwnd );  // This just ends up putting our window behind everything else
    if ( FAILED( result ) )
    {
      error = errAcquire;
      dic->Release();
      dic = 0;
      return;
    }
  }

  return;
}


// Constructor that acquires everything needed and fills in joystick vector

TJoySwitch::TJoySwitch( HINSTANCE  AppInstance, HWND  MainWindow )
{
  hinstance = AppInstance;
  hwnd      = MainWindow;

  // is_nt initialized later
  di  = 0;  // initialized further later
  dic = 0;  // initialized further later

  error = errNoInit;

  joystick.clear();  // redundant
  safeid = -1;       // illegal value
}


// Function that acquires everything needed and fills in joystick vector, return value indicates success
//  Do not call again, regardless of return value (use Timeout to retry if needed)

bool  TJoySwitch::Initialize( unsigned  MaxJoystickCount, DWORD  Timeout )
{
  try
  {
    joystick.resize( MaxJoystickCount );
  }
  catch ( ... )
  {
    error = errMemory;
    return  false;
  }

  DWORD  t_start, t_current;
  t_start = GetTickCount();
  do
  {
    initialize();
    t_current = GetTickCount();
    if ( t_current > t_start )  t_current = t_current - t_start;
    else                        t_current = t_start - t_current;
  }  while ( ( dic == 0 ) && ( t_current < Timeout) );
  if ( dic )  error = errNone;  // Clear any errors produced during earlier attempts to initialize, should be redundant
  else  return  false;

 // Enumerate using DInputJoyConfig (allows us to determine JoystickID exactly)
  DIJOYCONFIG  cfg;
  cfg.dwSize = sizeof( cfg );
  UINT  i;
  for ( i = 0; i < joystick.size(); ++i )
  {
    if ( SUCCEEDED( dic->GetConfig( i, &cfg, DIJC_GUIDINSTANCE | DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
    {
      joystick[ i ].ID = cfg.guidInstance;
    }
    else
    {
      joystick[ i ].ID = dummy;
    }
  }

 // Now enumerate using DInput (DInputJoyConfig finds old joysticks even on unused numbers)
  if ( FAILED( di->EnumDevices( DIDEVTYPE_JOYSTICK, JoystickCallback, static_cast< void* >( this ), DIEDFL_ALLDEVICES ) ) )  // not DIEDFL_ATTACHEDONLY, don't care if Windows thinks attached
  {
    error = errEnumeration;
  }

 // Go back and find a free one (any other errors to check for????) XXX
  for ( i = joystick.size(); i > 0; --i )
  {
    if ( ( DIERR_NOMOREITEMS != dic->GetConfig( i - 1, &cfg, DIJC_GUIDINSTANCE | DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) )
      && ( !joystick[ i - 1 ].Found ) )
    {
      safeid = i - 1;
      break;
    }
  }
  if ( safeid < 0 )
  {
    error = errNoEmpty;
    return  false;
  }
  else  return  true;
}


// Destructor that releases everything

TJoySwitch::~TJoySwitch()
{
  if ( dic )
  {
    dic->Unacquire();
    dic->Release();
  }

  if ( di )
  {
    di->Release();
    CoUninitialize();
  }
}


// Function to return error value

TJoySwitch::EError  TJoySwitch::GetError()
{
  return  error;
}


// Operator that returns the IDth joystick

TJoySwitch::TJoystick  TJoySwitch::operator[]( unsigned  ID )
{
  return  ( ID < joystick.size() ? joystick[ ID ] : TJoystick() );
}


// Function that swaps joystick at SourceID with joystick at DestID, return value indicates success
//  Method for NT/9x is from MSDN's online documentation for IDirectInputJoyConfig8

bool  TJoySwitch::Swap( unsigned  SourceID, unsigned  DestID )
{
  if ( SourceID == DestID )  return  true;
  if ( !dic )        return  false;
  if ( safeid < 0 )  return  false;
  if ( SourceID >= joystick.size() )  return  false;
  if ( DestID   >= joystick.size() )  return  false;

  DIJOYCONFIG  cfg;
  cfg.dwSize = sizeof( cfg );

 // Move SourceID -> safeid
  if ( FAILED( dic->GetConfig( SourceID, &cfg, DIJC_GUIDINSTANCE | DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
  {
    error = errGetConfig;
    return  false;
  }
  if ( is_nt )
  {
    if ( FAILED( dic->SetConfig( safeid, &cfg, DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
    {
      error = errSetConfig;
      return  false;
    }
    if ( FAILED( dic->DeleteConfig( SourceID ) ) )
    {
      error = errDeleteConfig;
      return  false;
    }
  }
  else  // Win9x/ME
  {
    if ( FAILED( dic->DeleteConfig( SourceID ) ) )
    {
      error = errDeleteConfig;
      return  false;
    }
    if ( FAILED( dic->SetConfig( safeid, &cfg, DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
    {
      error = errSetConfig;
      return  false;
    }
  }

 // Move DestID -> SourceID
  if ( FAILED( dic->GetConfig( DestID, &cfg, DIJC_GUIDINSTANCE | DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
  {
    error = errGetConfig;
    return  false;
  }
  if ( is_nt )
  {
    if ( FAILED( dic->SetConfig( SourceID, &cfg, DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
    {
      error = errSetConfig;
      return  false;
    }
    if ( FAILED( dic->DeleteConfig( DestID ) ) )
    {
      error = errDeleteConfig;
      return  false;
    }
  }
  else  // Win9x/ME
  {
    if ( FAILED( dic->DeleteConfig( DestID ) ) )
    {
      error = errDeleteConfig;
      return  false;
    }
    if ( FAILED( dic->SetConfig( SourceID, &cfg, DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
    {
      error = errSetConfig;
      return  false;
    }
  }

 // Move safeid -> DestID
  if ( FAILED( dic->GetConfig( safeid, &cfg, DIJC_GUIDINSTANCE | DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
  {
    error = errGetConfig;
    return  false;
  }
  if ( is_nt )
  {
    if ( FAILED( dic->SetConfig( DestID, &cfg, DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
    {
      error = errSetConfig;
      return  false;
    }
    if ( FAILED( dic->DeleteConfig( safeid ) ) )
    {
      error = errDeleteConfig;
      return  false;
    }
  }
  else  // Win9x/ME
  {
    if ( FAILED( dic->DeleteConfig( safeid ) ) )
    {
      error = errDeleteConfig;
      return  false;
    }
    if ( FAILED( dic->SetConfig( DestID, &cfg, DIJC_REGHWCONFIGTYPE | DIJC_GAIN | DIJC_CALLOUT ) ) )
    {
      error = errSetConfig;
      return  false;
    }
  }

  return  true;
}

